use std::collections::HashMap;

use serde::{Deserialize, Serialize};

#[derive(Clone, Debug, Serialize)]
pub(crate) struct WebhookEvent {
    pub topic: String,
    pub query: HashMap<String, String>,
    pub body: serde_json::Value,
}

#[derive(Clone, Debug, Serialize)]
pub(crate) struct ApiEvent {
    pub id: String,
    pub body: serde_json::Value,
    pub status: u16,
}

#[derive(Clone, Debug, Serialize)]
pub(crate) struct ErrorEvent {
    pub id: String,
    pub message: String,
}

#[derive(Clone, Debug)]
pub(crate) enum Event {
    Webhook(WebhookEvent),
    Api(ApiEvent),
    Error(ErrorEvent),
}

impl Event {
    pub fn error(id: String, message: impl ToString) -> Self {
        Self::Error(ErrorEvent {
            id,
            message: message.to_string(),
        })
    }
}

#[derive(Clone, Debug, serde_repr::Deserialize_repr, serde_repr::Serialize_repr)]
#[repr(u8)]
pub(crate) enum NtfyPriority {
    Min = 1,
    Low = 2,
    Default = 3,
    High = 4,
    Max = 5,
}

impl Default for NtfyPriority {
    fn default() -> Self {
        Self::Default
    }
}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub(crate) struct NtfyAction {
    pub action: String,
    pub label: String,
    pub url: String,
}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub(crate) struct NtfyMessage {
    pub topic: String,
    pub message: String,
    pub title: Option<String>,
    #[serde(default, skip_serializing_if = "Vec::is_empty")]
    pub tags: Vec<String>,
    #[serde(default)]
    pub priority: NtfyPriority,
    #[serde(default)]
    pub click: Option<String>,
    #[serde(default, skip_serializing_if = "Vec::is_empty")]
    pub actions: Vec<NtfyAction>,
    #[serde(default)]
    pub markdown: bool,
    #[serde(default)]
    pub icon: Option<String>,
    #[serde(default)]
    pub delay: Option<String>,
}

#[derive(Clone, Debug)]
pub(crate) enum Message {
    Ntfy(NtfyMessage),
}

#[derive(Clone, Debug, Deserialize)]
pub(crate) enum Method {
    Get,
    Post,
}

#[derive(Clone, Debug, Deserialize)]
pub(crate) struct ApiTask {
    pub id: String,
    pub interval: u64,
    pub method: Method,
    pub url: String,
    pub body: serde_json::Value,
    pub query: HashMap<String, String>,
}
