use std::sync::mpsc::Sender;
use std::sync::Arc;
use std::time::Duration;

use color_eyre::eyre::Context as _;
use color_eyre::Result;
use reqwest::Client;
use tokio::runtime;
use tokio::sync::mpsc::UnboundedReceiver;
use tokio::sync::Mutex;
use tokio_stream::wrappers::IntervalStream;
use tokio_stream::{StreamExt as _, StreamMap};

use crate::types::{ApiEvent, ApiTask, Event, Method};

async fn perform_request(client: &Client, task: &ApiTask) -> Result<Event> {
    let req = match task.method {
        Method::Get => client.get(&task.url),
        Method::Post => {
            let body = serde_json::to_vec(&task.body)
                .expect("Type `serde_json::Value` should always be serializable");
            client.post(&task.url).body(body)
        }
    };
    let res = req.query(&task.query).send().await?;
    let status = res.status().as_u16();

    let body: serde_json::Value = res.json().await?;

    let event = ApiEvent {
        id: task.id.clone(),
        body,
        status,
    };
    Ok(Event::Api(event))
}

#[tracing::instrument(skip_all)]
pub async fn run(mut api_rx: UnboundedReceiver<ApiTask>, event_tx: Sender<Event>) -> Result<()> {
    let tasks = Arc::new(Mutex::new(StreamMap::new()));

    let client = Client::builder()
        .build()
        .wrap_err("Failed to build HTTP client")?;

    tokio::spawn({
        let tasks = tasks.clone();
        async move {
            while let Some(task) = api_rx.recv().await {
                tracing::trace!("Received new API task: {:?}", task);

                let id = task.id.clone();
                let interval = tokio::time::interval(Duration::from_secs(task.interval));

                let task = Arc::new(task);
                let mut tasks = tasks.lock().await;
                tasks.insert(id, IntervalStream::new(interval).map(move |_| task.clone()));
            }
            tracing::error!("API task channel closed");
        }
    });

    loop {
        // We need to guarantee that the lock on `tasks` is released as soon as possible.
        // To ensure that it isn't held for the entirety of the `sleep`, the indirection
        // via the extra `bool` value is used, so that the lock can be dropped immediately,
        // and during the `sleep` new streams can be created.
        let wait = {
            let mut tasks = tasks.lock().await;
            match tasks.next().await {
                Some((_, task)) => {
                    let event = perform_request(&client, &task)
                        .await
                        .wrap_err("Failed to perform API request")?;
                    event_tx.send(event).wrap_err("Failed to send event")?;

                    false
                }
                None => true,
            }
        };

        if wait {
            // Ideally we would be able to wait explicitly for the first stream
            // to be registered. But as a workaround, we have to idle wait.
            tokio::time::sleep(Duration::from_millis(500)).await;
        }
    }
}

#[tracing::instrument("api::worker", skip_all)]
pub fn worker(api_rx: UnboundedReceiver<ApiTask>, event_tx: Sender<Event>) -> Result<()> {
    let rt = runtime::Builder::new_current_thread()
        .thread_name("api")
        .enable_io()
        .enable_time()
        .build()?;

    rt.block_on(run(api_rx, event_tx))
}
