use std::sync::mpsc::Sender;

use color_eyre::eyre::{self, Context as _};
use color_eyre::Result;
use reqwest::{header, Client};
use tokio::runtime;
use tokio::sync::mpsc::UnboundedReceiver;

use crate::types::{Event, Message, NtfyMessage};

static APP_USER_AGENT: &str = concat!(env!("CARGO_PKG_NAME"), "/", env!("CARGO_PKG_VERSION"),);

#[tracing::instrument]
async fn send_ntfy(client: &Client, url: &String, data: &NtfyMessage) -> Result<()> {
    let body = serde_json::to_string(data)?;
    tracing::trace!("JSON: {}", body);
    let res = client.post(url).body(body).send().await?;
    if res.status().as_u16() >= 400 {
        let body = res.text().await?;
        eyre::bail!("Ntfy server returned error: {}", body);
    }

    Ok(())
}

#[tracing::instrument(skip_all)]
pub async fn run(event_tx: Sender<Event>, mut ntfy_rx: UnboundedReceiver<Message>) -> Result<()> {
    let ntfy_url = std::env::var("NTFY_URL").wrap_err("Missing env var 'NTFY_URL'")?;
    let ntfy_token = std::env::var("NTFY_TOKEN").wrap_err("Missing env var 'NTFY_TOKEN'")?;

    let ntfy_client = Client::builder()
        .default_headers({
            let mut headers = header::HeaderMap::new();

            let auth = format!("Bearer {}", ntfy_token);
            let mut auth = header::HeaderValue::from_str(&auth)?;
            auth.set_sensitive(true);

            headers.insert(header::AUTHORIZATION, auth);

            headers
        })
        .user_agent(APP_USER_AGENT)
        .build()?;

    while let Some(message) = ntfy_rx.recv().await {
        tracing::trace!("Received notification: {:?}", message);

        match message {
            Message::Ntfy(data) => {
                if let Err(err) = send_ntfy(&ntfy_client, &ntfy_url, &data).await {
                    tracing::error!("Failed to send to Ntfy: {:?}", err);
                    event_tx.send(Event::error(data.topic, err))?;
                }
            }
        }
    }

    tracing::debug!("Stopped receiving messages");
    Ok(())
}

#[tracing::instrument(skip_all)]
pub fn worker(event_tx: Sender<Event>, ntfy_rx: UnboundedReceiver<Message>) -> Result<()> {
    let rt = runtime::Builder::new_current_thread()
        .thread_name("sender")
        .enable_io()
        .build()?;

    rt.block_on(run(event_tx, ntfy_rx))
}
